/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/


#include "AMRNavierStokesHeaders.H"






// this function manages the pressure initialization after
// initialization and regridding 
void
AMRNavierStokes::initializeGlobalPressure()
{

  if (s_verbosity >= 3) {
    pout() << "AMRNavierStokes::initializeGlobalPressure " 
	   << m_level << endl;
  }

  // index through levels to find out what finest level is
  // (there must be a better way to do this!)
  AMRNavierStokes* thisNSPtr = this;
  while (!thisNSPtr->finestLevel()) {
    thisNSPtr = thisNSPtr->finerNSPtr();
  }
  assert(thisNSPtr->finestLevel());
  int finest_level = thisNSPtr->m_level;
  
  // first compute dtInit
  Real cur_time = m_time;
  // also save dt for each level so it can be reset later
  Vector<Real> dtSave(finest_level+1, 1.0e8);
  Real dtLevel;
  Real dtInit = 10000000.0;
  thisNSPtr = this;
  for (int lev=m_level; lev<=finest_level; lev++) {
    dtSave[lev] = thisNSPtr->m_dt;
    dtLevel = thisNSPtr->computeDt();
    if (dtLevel < dtInit) dtInit = dtLevel;
    thisNSPtr = thisNSPtr->finerNSPtr();
  }
  dtInit *= 0.5;

  // pressures should already be initialized to first
  // guess (most likely 0)
  
  // now loop through levels and compute estimate of pressure
  for (int iter=0; iter<s_num_init_passes; iter++) {
    int lbase = m_level;
    // is this still necessary?    
    //if (m_level > 0) lbase = lbase-1;

    AMRNavierStokes* thisNSPtr = this;
    //if (m_level > 0) thisNSPtr = crseNSPtr();

    for (int lev=lbase; lev<=finest_level; lev++) {
      thisNSPtr->initializeLevelPressure(cur_time, dtInit);
      thisNSPtr = thisNSPtr->finerNSPtr();
    }

    // now reset times and dt
    thisNSPtr = this;
    for (int lev=lbase; lev<=finest_level; lev++) {
      thisNSPtr->resetStates(cur_time);
      thisNSPtr->dt(dtSave[lev]);
      thisNSPtr = thisNSPtr->finerNSPtr();
    }
    
  } // end loop over init passes
  
}


// this function does all the level-based operations for 
// initializing the pressure
void
AMRNavierStokes::initializeLevelPressure(Real a_currentTime, Real a_dtInit)
{

  if (s_verbosity >= 3) {
    pout() << "AMRNavierStokes::initializeLevelPressure " 
	   << m_level << endl;
  }
  
  
  Real new_time = a_currentTime + a_dtInit;
  m_dt = a_dtInit;
  swapOldAndNewStates();
  
  const DisjointBoxLayout levelGrids = newVel().getBoxes();
  
  // just to be sure, initialize flux registers
  if (!finestLevel()) {
    m_flux_register.setToZero();
    m_lambda_flux_reg.setToZero();
    for (int comp=0; comp<m_scal_fluxreg_ptrs.size(); comp++) {
      m_scal_fluxreg_ptrs[comp]->setToZero();
    }
  }

  // compute advection velocities
  LevelData<FluxBox> advectionVel(levelGrids,1);
  
  if (s_set_bogus_values) {
    DataIterator dit = advectionVel.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      advectionVel[dit()].setVal(s_bogus_value);
    }
  }
  
  computeAdvectionVelocities(advectionVel);
  
  // since this is just for initialization, no need to update
  // advected scalars.  
  
  // now predict velocities -- this returns the advection
  // term.  put this in new_vel.
  LevelData<FArrayBox>& uStar = newVel();
  predictVelocities(uStar, advectionVel);
  
  computeUStar(uStar);
  
  // if a coarser level exists, will need coarse-level data for proj
  LevelData<FArrayBox>* crseVelPtr = NULL;
  
  if (m_level > 0) {
    const DisjointBoxLayout& crseGrids = crseNSPtr()->newVel().getBoxes();
    crseVelPtr = new LevelData<FArrayBox>(crseGrids, SpaceDim);
    // coarse velocity BC data is interpolated in time
    crseNSPtr()->fillVelocity(*crseVelPtr, new_time);
  }
  
  // need to do physical boundary conditions and exchanges
  bool isViscous = (s_nu > 0);
  basicVelBC* velBCPtr = m_physBCPtr->uStarBC(isViscous);

  velBCPtr->applyBCs(uStar, problemDomain(), m_dx);
  Interval velComps(0,SpaceDim-1);
  delete velBCPtr;
  
  // (Level projection calls exchange)
  //uStar.exchange(velComps);

  m_projection.LevelProject(uStar, crseVelPtr, new_time, m_dt);
  
  // as things stand now, physical BC's are reset in Level Projection
  
  // clean up temp storage
  if (crseVelPtr != NULL) {
    delete crseVelPtr;
    crseVelPtr = NULL;
  }
  
  
}


void
AMRNavierStokes::smoothVelocityField(int a_lbase)
{
  // in this function, we take the filpatched fields stored in 
  // the old-time storage (which have been modified to contain
  // s - mu*lap(s) in the regrid() function and perform an 
  // elliptic solve which should result in a smoothed s

  // first need to loop over levels and put together amr storage
  // this should be called on the lbase level 
  assert (m_level == a_lbase);
  assert (m_regrid_smoothing_done);  

  AMRNavierStokes* thisNSPtr = this;
  while (!thisNSPtr->finestLevel()) {
    thisNSPtr = thisNSPtr->finerNSPtr();
  }
  assert (thisNSPtr->finestLevel());
  int finest_level = thisNSPtr->m_level;

  thisNSPtr = this;
  int startLev = m_level;
  if (m_level > 0) {
    startLev = m_level-1;
    thisNSPtr = thisNSPtr->crseNSPtr();
  }

  // now set up multilevel stuff
  Vector<LevelData<FArrayBox>*> oldS(finest_level+1, NULL);
  Vector<LevelData<FArrayBox>*> newS(finest_level+1, NULL);
  Vector<DisjointBoxLayout> amrGrids(finest_level+1);
  Vector<int> amrRefRatios(finest_level+1,0);
  Vector<Real> amrDx(finest_level+1,0);
  Vector<ProblemDomain> amrDomains(finest_level+1);
  // also will need to avg down new stuff
  Vector<CoarseAverage*> amrAvgDown(finest_level+1,NULL);

  // this is debugging stuff...
  Vector<LevelData<FArrayBox>*> beforeAfterData(finest_level+1, NULL);
  // this will contain before, after, diff, Lap(before), Lap(after), Lap(diff)
  int numDebugComp= 6*SpaceDim;

  // loop over levels, allocate temp storage for velocities,
  // set up for amrsolves
  for (int lev=startLev; lev<=finest_level; lev++) {
    const DisjointBoxLayout& levelGrids = thisNSPtr->newVel().getBoxes();
    // since AMRSolver can only handle one component at a 
    // time, need to allocate temp space to copy stuff
    // into, and then back out of to compute laplacian
    IntVect ghostVect(D_DECL(1,1,1));
    newS[lev] = new LevelData<FArrayBox>(levelGrids, 1, ghostVect);
    oldS[lev] = new LevelData<FArrayBox>(levelGrids,1,ghostVect);
    if (s_dump_smoothing_plots) 
      {
        beforeAfterData[lev] = new LevelData<FArrayBox>(levelGrids, 
                                                        numDebugComp,
                                                        IntVect::TheZeroVector());
    
        // while we're here, copy "before" velocity field into 
        // debugging storage
        Interval velComps(0,SpaceDim-1);
        thisNSPtr->newVel().copyTo(velComps, *beforeAfterData[lev], velComps);
        /// also save this for the differencing
        Interval diffComps(2*SpaceDim, 3*SpaceDim-1);
        thisNSPtr->newVel().copyTo(velComps, *beforeAfterData[lev], diffComps);
      }  // end setup for debugging dump

    amrGrids[lev] = levelGrids;
    amrRefRatios[lev] = thisNSPtr->refRatio();
    amrDx[lev] = thisNSPtr->Dx();
    amrDomains[lev] = thisNSPtr->problemDomain();
    thisNSPtr = thisNSPtr->finerNSPtr();
    if (lev>startLev) {
      amrAvgDown[lev] = new CoarseAverage(levelGrids,1, 
					  amrRefRatios[lev-1]);
    }
  }
  
  AMRSolver tempSolver;
  // allocate solver -- do this in a separate function in
  // order to make sure the two steps which involve
  // elliptic operators are consistent.
  defineRegridAMRSolver(tempSolver, amrGrids, 
                        amrDomains, amrDx, 
                        amrRefRatios, m_level);
  
  
  // now loop over velocity components
  if (s_nu > 0) {
    for (int dir=0; dir<SpaceDim; ++dir) {
      Interval velComps(dir,dir);
      Interval tempComps(0,0);
      thisNSPtr = this;
      if (startLev<m_level) thisNSPtr = thisNSPtr->crseNSPtr();
      for (int lev=startLev; lev<=finest_level; lev++) {
	thisNSPtr->oldVel().copyTo(velComps, *oldS[lev], tempComps);
	// do this as initial guess?
	thisNSPtr->oldVel().copyTo(velComps, *newS[lev], tempComps);
	thisNSPtr = thisNSPtr->finerNSPtr();
      }
      
      // now do elliptic solve
      tempSolver.solveAMR(newS, oldS);

      // average new s down to invalid regions
      for (int lev=finest_level; lev>startLev; lev--) {
	amrAvgDown[lev]->averageToCoarse(*newS[lev-1], *newS[lev]);
      }

      // now copy back to new velocity field
      thisNSPtr = this;
      for (int lev=m_level; lev <= finest_level; lev++) {
	newS[lev]->copyTo(tempComps, thisNSPtr->newVel(), velComps);
        
        if (s_dump_smoothing_plots) 
          {
            // copy to before-after debugging storage
            int copyComp = SpaceDim+dir;
            Interval destInterval(copyComp, copyComp);
            LevelData<FArrayBox>& debugLDF = *beforeAfterData[lev];
            newS[lev]->copyTo(tempComps, debugLDF, destInterval);
            // this is also a good time to compute difference
            DataIterator levelDit = beforeAfterData[lev]->dataIterator();
            int minusComp = copyComp+SpaceDim;
            
            for (levelDit.begin(); levelDit.ok(); ++levelDit) 
              {
                FArrayBox& debugFab = debugLDF[levelDit()];
                debugFab.minus(debugFab, copyComp, minusComp, 1);
              }
          } // end before-after debugging stuff
      
	thisNSPtr = thisNSPtr->finerNSPtr();
      }
    } // end loop over velocity components

    if (s_dump_smoothing_plots) 
      {
        // first compute laplacian of difference
        PoissonOp poisOp;
        // for now, just use extrap BC's, since they're probably the safest
        DomainGhostBC tempBC;
        SideIterator sit;
        for (int dir=0; dir<SpaceDim; dir++) {
          for (sit.reset(); sit.ok(); ++sit) {
            ExtrapBC thisBC(dir,sit());
            tempBC.setBoxGhostBC(thisBC);
          }
        }
        
        poisOp.setDomainGhostBC(tempBC);

        AMRSolver poissSolver(amrGrids, amrDomains, amrDx, amrRefRatios,
                              amrDomains.size(), m_level, &poisOp);

        for (int dir=0; dir<SpaceDim; dir++) 
          {
            int oldVelComp = dir;
            int newVelComp = SpaceDim+dir;
            int diffComp = 2*SpaceDim + dir;
            int oldLapComp = 3*SpaceDim+dir;
            int newLapComp = 4*SpaceDim+dir;
            int diffLapComp = 5*SpaceDim+dir;
            Interval tempInterval(0,0);
            Interval oldVelInterval(oldVelComp, oldVelComp);
            Interval newVelInterval(newVelComp, newVelComp);
            Interval diffInterval(diffComp, diffComp);
            Interval oldLapInterval(oldLapComp, oldLapComp);
            Interval newLapInterval(newLapComp, newLapComp);
            Interval diffLapInterval(diffLapComp, diffLapComp);

            // first compute Lap(pre-smoothing vel)
            for (int lev= startLev; lev<= finest_level; lev++) 
              {
                // since amrsolver can only handle one component
                // at a time, copy into single-component storage
                beforeAfterData[lev]->copyTo(oldVelInterval, 
                                             *oldS[lev], 
                                             tempInterval);
              }
            // now loop over levels, compute Lap(oldVel)
            for (int lev= m_level; lev<= finest_level; lev++)
              {
                poissSolver.applyAMROperator(oldS, *newS[lev], lev);
                if (lev>m_level)
                  {
                    // do averaging down
                    amrAvgDown[lev]->averageToCoarse(*newS[lev-1],
                                                     *newS[lev]);
                  }
              }

            // now copy into dump data storage
            for (int lev= m_level; lev<=finest_level; lev++) 
              {
                LevelData<FArrayBox>& debugLDF = *beforeAfterData[lev];
                newS[lev]->copyTo(tempInterval, debugLDF, oldLapInterval);
              }
          
            // now do Lap(post-smoothing vel)
            for (int lev=startLev; lev<= finest_level; lev++) 
              {
                beforeAfterData[lev]->copyTo(newVelInterval,
                                             *oldS[lev], 
                                             tempInterval);
              }

            // now loop over levels, compute Lap(newVel)
            for (int lev=m_level; lev<=finest_level; lev++) 
              {
                poissSolver.applyAMROperator(oldS, *newS[lev], lev);
                if (lev>m_level)
                  {
                    amrAvgDown[lev]->averageToCoarse(*newS[lev-1],
                                                     *newS[lev]);
                  }
              }

            // now copy into dump data storage
            for (int lev=m_level; lev<=finest_level; lev++)
              {
                LevelData<FArrayBox>& debugLDF = *beforeAfterData[lev];
                newS[lev]->copyTo(tempInterval, debugLDF, newLapInterval);
              }


            // now compute Lap(difference)
            if (m_level > 0) 
              {
                // level less than this level had no change
                LevelData<FArrayBox>& lBaseDiff = *oldS[m_level-1];
                DataIterator baseDit = lBaseDiff.dataIterator();
                for (baseDit.begin(); baseDit.ok(); ++baseDit)
                  {
                    lBaseDiff[baseDit()].setVal(0.0);
                  }
              } // end setting unchange level's diff to 0

            for (int lev= m_level; lev<= finest_level; lev++) 
              {
                // since amrsolver can only handle one component
                // at a time, copy into single-component storage
                LevelData<FArrayBox>& debugLDF = *beforeAfterData[lev];
                debugLDF.copyTo(diffInterval, *oldS[lev], tempInterval);
              }
            // now loop over levels, compute Lap(diff(vel))
            for (int lev=m_level; lev<=finest_level; lev++) 
              {
                poissSolver.applyAMROperator(oldS, *newS[lev], lev);
                if (lev > m_level) 
                  {
                    // do averaging down
                    amrAvgDown[lev]->averageToCoarse(*newS[lev-1], 
                                                     *newS[lev]);
                  }
              }
            // now copy into dumpData
            for (int lev=m_level; lev<=finest_level; lev++) 
              {
                LevelData<FArrayBox>& debugLDF = *beforeAfterData[lev];
                newS[lev]->copyTo(tempInterval, debugLDF, diffLapInterval);
              }
          } // end loop over velocity components


        // do what we need to do to dump this out
        char iter_str[100];
        sprintf(iter_str, "%s%d.%d.hdf5", "regridSmooth", a_lbase, 
                m_level_steps);
        string fname(iter_str);
        
        Vector<string>  dataNames(beforeAfterData[a_lbase]->nComp());
        int nameComp=0;
        D_TERM(
               dataNames[nameComp] = "oldXvel"; nameComp++; ,
               dataNames[nameComp] = "oldYvel"; nameComp++; , 
               dataNames[nameComp] = "oldZvel"; nameComp++;);
        
        D_TERM(
               dataNames[nameComp] = "newXvel"; nameComp++; ,
               dataNames[nameComp] = "newYvel"; nameComp++; ,
               dataNames[nameComp] = "newZvel"; nameComp++; );
        
        D_TERM(
               dataNames[nameComp] = "diffXvel"; nameComp++; ,
               dataNames[nameComp] = "diffYvel"; nameComp++; ,
               dataNames[nameComp] = "diffZvel"; nameComp++; );


        D_TERM(
               dataNames[nameComp] = "LapOldXvel"; nameComp++; ,
               dataNames[nameComp] = "LapOldYvel"; nameComp++; ,
               dataNames[nameComp] = "LapOldZvel"; nameComp++; );

        D_TERM(
               dataNames[nameComp] = "LapNewXvel"; nameComp++; ,
               dataNames[nameComp] = "LapNewYvel"; nameComp++; ,
               dataNames[nameComp] = "LapNewZvel"; nameComp++; );

        D_TERM(
               dataNames[nameComp] = "LapDiffXvel"; nameComp++; ,
               dataNames[nameComp] = "LapDiffYvel"; nameComp++; ,
               dataNames[nameComp] = "LapDiffZvel"; nameComp++; );
        
        Interval levelsToDump(a_lbase, finest_level);
        WritePartialAMRHierarchyHDF5(fname, 
                                     amrGrids, 
                                     beforeAfterData,
                                     dataNames,
                                     amrDomains[a_lbase].domainBox(),
                                     amrDx[a_lbase],
                                     m_dt, m_time, amrRefRatios,
                                     levelsToDump);
      } // end dump smoothing info

  }// end if viscous

  // since scalars won't need temp storge, clean it up here
  for (int lev=startLev; lev<= finest_level; lev++) {
    if (newS[lev] != NULL) {
      delete newS[lev];
      newS[lev] = NULL;
    } 
    if (oldS[lev] != NULL) {
      delete oldS[lev];
      oldS[lev] = NULL;
    }
    if (beforeAfterData[lev] != NULL) 
      {
        delete beforeAfterData[lev];
        beforeAfterData[lev] = NULL;
      }
  }
  
  // now do scalars
  for (int scalComp=0; scalComp < s_num_scal_comps; scalComp++) {
    // only do this if scalar is diffused
    if (s_scal_coeffs[scalComp] > 0) {
      thisNSPtr =this;
      for (int lev=startLev; lev <= finest_level; lev++) {
	newS[lev] = thisNSPtr->m_scal_new[scalComp];
	oldS[lev] = thisNSPtr->m_scal_old[scalComp];
	thisNSPtr = thisNSPtr->finerNSPtr();
      }

      tempSolver.solveAMR(newS, oldS);

      // now do averaging down
      for (int lev=finest_level; lev>m_level; lev--) {
	amrAvgDown[lev]->averageToCoarse(*newS[lev-1], *newS[lev]);
      }
    }  // end if this scalar is diffused

  }  // end loop over scalars

  // finally, loop over levels, clean up storage, and reset boolean 
  thisNSPtr = this;
  for (int lev=m_level; lev<= finest_level; lev++) {
    if (amrAvgDown[lev] != NULL) {
      delete amrAvgDown[lev];
      amrAvgDown[lev] = NULL;
    }
    thisNSPtr->m_regrid_smoothing_done = false;
    thisNSPtr = thisNSPtr->finerNSPtr();
  }


}



void
AMRNavierStokes::defineRegridAMRSolver(AMRSolver& a_solver, 
                                       const Vector<DisjointBoxLayout>& a_grids,
                                       const Vector<ProblemDomain>& a_domains,
                                       const Vector<Real>& a_amrDx,
                                       const Vector<int>& a_refRatios,
                                       const int& a_lBase)
{
  

  HelmholtzOp helmOp;


  // for now, just use extrap BC's, since they're probably the safest
  DomainGhostBC tempBC;
  SideIterator sit;
  for (int dir=0; dir<SpaceDim; dir++) {
    for (sit.reset(); sit.ok(); ++sit) {
      ExtrapBC thisBC(dir,sit());
      tempBC.setBoxGhostBC(thisBC);
    }
  }

  helmOp.setDomainGhostBC(tempBC);

  // want to use dt from lbase
  Real dtLBase = m_dt;
  if (a_lBase > m_level) {
    AMRNavierStokes* thisNSPtr = finerNSPtr();
    while (a_lBase > thisNSPtr->m_level) {
      thisNSPtr = thisNSPtr->finerNSPtr();
    }
    dtLBase = thisNSPtr->m_dt;
  } else if (a_lBase < m_level) {
    AMRNavierStokes* thisNSPtr = crseNSPtr();
    while (a_lBase < thisNSPtr->m_level) {
      thisNSPtr = thisNSPtr->crseNSPtr();
    }
    dtLBase = thisNSPtr->m_dt;
  }


  // define coefficient
  Real mu = -s_regrid_smoothing_coeff*dtLBase*s_nu;
  helmOp.setHelmCoeff(mu);

  a_solver.define(a_grids, a_domains, a_amrDx, a_refRatios,
		  a_domains.size(), a_lBase, &helmOp);

}




