/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/


#include "AMRNavierStokesHeaders.H"



Real
AMRNavierStokes::computeInitialDt() 
{
  if (s_verbosity >= 3) {
    pout() << "AMRNavierStokes::computeInitialDt " << m_level << endl;
  }

  return s_init_shrink*computeDt();
}

Real
AMRNavierStokes::computeDt()
{
  if (s_verbosity >= 3) {
    pout() << "AMRNavierStokes::computeDt " << m_level << endl;
  }

  Real dt; 

  // if prescribedDt > 0 then use that
  // (don't worry about not being on level 0 -- Amr class
  // will take care of that
  if (s_prescribedDt > 0.0) {
    dt = s_prescribedDt;
  } else {
    
    Real MaxVel = 0;
    
    Interval velComps(0,SpaceDim-1);
    // compute maximum velocity  (MaxNorm)
    MaxVel = norm(newVel(), velComps, 0);
    
    
    if (MaxVel != 0) {
      dt = m_cfl*m_dx/MaxVel;
    } else {
      dt = s_max_dt;
    }
    
    
    // check to see if boundary conditions restrict dt
    m_physBCPtr->computeBoundaryDt(dt, m_cfl, m_dx);
  }
  
  return dt;
}


void 
AMRNavierStokes::CFL(Real a_cfl)
{ 
  m_cfl = a_cfl;
}

void
AMRNavierStokes::refinementThreshold(Real a_thresh) 
{ 
  m_refine_threshold = a_thresh;
}
    
bool
AMRNavierStokes::finestLevel() const
{
  return m_finest_level;
}

bool
AMRNavierStokes::isEmpty() const
{
  return m_is_empty;
}

void
AMRNavierStokes::finestLevel(bool a_finest_level) 
{
  m_finest_level = a_finest_level;
  m_projection.isFinestLevel(a_finest_level);
}

Real
AMRNavierStokes::Dx() const 
{
  return m_dx;
}


AMRNavierStokes* 
AMRNavierStokes::finerNSPtr() const
{
  
  AMRNavierStokes* fine_e_ptr;
  if (m_finer_level_ptr != NULL) {
    fine_e_ptr = dynamic_cast <AMRNavierStokes*> (m_finer_level_ptr);
    if (fine_e_ptr == NULL) {
      MayDay::Error ("AMRNavierStokes::fineNSPtr -- fineptr not castable to AMRNavierStokes*");
    }
  } else {
    fine_e_ptr = NULL;
  }
  

  return fine_e_ptr;
}

AMRNavierStokes* 
AMRNavierStokes::crseNSPtr() const
{
  
  AMRNavierStokes* crse_e_ptr;
  if (m_coarser_level_ptr != NULL) {
    crse_e_ptr = dynamic_cast <AMRNavierStokes*> (m_coarser_level_ptr);
    if (crse_e_ptr == NULL) {
      MayDay::Error ("AMRNavierStokes::crseNSPtr -- crseptr not castable to AMRNavierStokes*");
    }
  } else {
    crse_e_ptr = NULL;
  }
  

  return crse_e_ptr;
}


LevelData<FArrayBox>& 
AMRNavierStokes::newVel()
{
  assert(m_vel_new_ptr != NULL);

  return *(m_vel_new_ptr);
}

LevelData<FArrayBox>& 
AMRNavierStokes::oldVel()
{
  assert(m_vel_old_ptr != NULL);

  return *(m_vel_old_ptr);
}


LevelData<FArrayBox>& 
AMRNavierStokes::newLambda()
{
  assert(m_lambda_new_ptr != NULL);

  return *(m_lambda_new_ptr);
}


LevelData<FArrayBox>& 
AMRNavierStokes::oldLambda()
{
  assert(m_lambda_old_ptr != NULL);

  return *(m_lambda_old_ptr);
}


LevelData<FArrayBox>& 
AMRNavierStokes::newScal(const int a_comp)
{
  assert(m_scal_new[a_comp] != NULL);

  return *(m_scal_new[a_comp]);
}


LevelData<FArrayBox>& 
AMRNavierStokes::oldScal(const int a_comp)
{
  assert(m_scal_old[a_comp] != NULL);

  return *(m_scal_old[a_comp]);
}




const LevelData<FArrayBox>& 
AMRNavierStokes::newVel() const
{
  assert(m_vel_new_ptr != NULL);

  return *(m_vel_new_ptr);
}

const LevelData<FArrayBox>& 
AMRNavierStokes::oldVel() const
{
  assert(m_vel_old_ptr != NULL);

  return *(m_vel_old_ptr);
}

const LevelData<FArrayBox>& 
AMRNavierStokes::newLambda() const 
{
  assert(m_lambda_new_ptr != NULL);

  return *(m_lambda_new_ptr);
}


const LevelData<FArrayBox>& 
AMRNavierStokes::oldLambda() const 
{
  assert(m_lambda_old_ptr != NULL);

  return *(m_lambda_old_ptr);
}


const LevelData<FArrayBox>& 
AMRNavierStokes::newScal(const int a_comp) const 
{
  assert(m_scal_new[a_comp] != NULL);

  return *(m_scal_new[a_comp]);
}


const LevelData<FArrayBox>& 
AMRNavierStokes::oldScal(const int a_comp) const 
{
  assert(m_scal_old[a_comp] != NULL);

  return *(m_scal_old[a_comp]);
}

void
AMRNavierStokes::setPhysBC(const PhysBCUtil& a_BC)
{
  m_physBCPtr = a_BC.newPhysBCUtil();
}

PhysBCUtil*
AMRNavierStokes::getPhysBCPtr() const
{
  return m_physBCPtr;
}



// this also changes time levels
void
AMRNavierStokes::swapOldAndNewStates() 
{
  LevelData<FArrayBox>* tempPtr;
  
  // swap both velocity and scalar data
  tempPtr = m_vel_new_ptr;
  m_vel_new_ptr = m_vel_old_ptr;
  m_vel_old_ptr = tempPtr;

  tempPtr = m_lambda_new_ptr;
  m_lambda_new_ptr = m_lambda_old_ptr;
  m_lambda_old_ptr = tempPtr;

  // loop over scalar components
  assert (m_scal_new.size() == m_scal_old.size());

  for (int comp=0; comp<m_scal_new.size(); comp++) {
    tempPtr = m_scal_new[comp];
    m_scal_new[comp] = m_scal_old[comp];
    m_scal_old[comp] = tempPtr;
  }
  
  time(m_time + m_dt);
}

// kinda like swapOldAndNewStates, but resets time to a_time
void
AMRNavierStokes::resetStates(const Real a_time)
{
  LevelData<FArrayBox>* tempPtr;
  tempPtr = m_vel_new_ptr;
  m_vel_new_ptr = m_vel_old_ptr;
  m_vel_old_ptr = tempPtr;

  tempPtr = m_lambda_new_ptr;
  m_lambda_new_ptr = m_lambda_old_ptr;
  m_lambda_old_ptr = tempPtr;

  // loop over scalar components
  assert (m_scal_new.size() == m_scal_old.size());
  
  for (int comp=0; comp<m_scal_new.size(); comp++) {
    tempPtr = m_scal_new[comp];
    m_scal_new[comp] = m_scal_old[comp];
    m_scal_old[comp] = tempPtr;
  }

  time(a_time);
}
