/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// VelocityPoissonOp.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/

#include "VelocityPoissonOp.H"
#include "PoissonOpF_F.H"

#include "Timer.H"

//timercomment//extern Timer NST_Everything;
//timercomment//Timer TF_OPERATORLAP_VelocityPoissonOp ("VPoi OPERATORLAP",  NST_Everything, DIAG_TABLE4);

LevelOp* 
VelocityPoissonOp::new_levelop() const
{
  //only check to see if the boundary conditions are
  //defined.  not necessary to have whole thing defined.
  //the boundary conditions need to be defined because
  //the interface does not know about the boundary
  //condtions.  the solvers call the other define functions
  //in our inimitable two-stage construction
  assert(m_isBCDefined);
  VelocityPoissonOp* newop = new VelocityPoissonOp();
  if(newop == NULL)
    {
      MayDay::Error("Out of Memory in PoissonOp::new_levelop");
    }
  newop->setDomainGhostBC(m_domghostbc);
  if (m_bottom_smoother_ptr != NULL)
    {
      newop->setBottomSmoother(*m_bottom_smoother_ptr);
    }
  return static_cast<LevelOp*>(newop);
}


/***********************/
// default constructor
/***********************/
VelocityPoissonOp::VelocityPoissonOp():PoissonOp()
{
}


// destructor
VelocityPoissonOp::~VelocityPoissonOp()
{
}

/***********************/
// evaluate operator, inhomogeneous C/F boundary conditions
// note that operator applies C/F interpolation and copy bc's
/***********************/
void
VelocityPoissonOp::applyOpI(LevelData<FArrayBox>& a_phi, 
                    const LevelData<FArrayBox>* a_phicPtr, 
                    LevelData<FArrayBox>& a_lofPhi,
		    Vector<DomainGhostBC>& a_bcVect)
{
  // don't call PoissonOp's isDefined function because
  // we don't care if BC is defined.
  assert(m_isDefined);
  assert(m_ihcfiEnabled);
  assert (a_phi.nComp() == a_lofPhi.nComp());
  assert (a_phi.nComp() == m_ncomp);

  if(a_phicPtr != NULL)
    {
      // apply C/F boundary conditions...
      CFInterp(a_phi,*a_phicPtr);
    }

  //fill in intersection of ghostcells and phi's boxes
  a_phi.exchange(a_phi.interval(), m_exchangeCopier);
    

  Real dx = m_dxLevel;
    
  DataIterator dit = a_phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {

#ifndef NDEBUG
      FArrayBox& thisPhi = a_phi[dit()];
      FArrayBox& thisLPhi = a_lofPhi[dit()];
      const Box& thisBox = m_grids.get(dit());
      assert(thisPhi.box().contains(grow(thisBox,1)));
      assert(thisLPhi.box().contains(thisBox));
#endif

      for (int comp=0; comp<a_bcVect.size(); comp++) {
	a_bcVect[comp].applyInhomogeneousBCs(a_phi[dit()], 
					     m_domain, 
					     m_dxLevel);
      }
      
      //timercomment//TF_OPERATORLAP_VelocityPoissonOp.start();
      FORT_OPERATORLAP(CHF_FRA(a_lofPhi[dit()]),
                       CHF_CONST_FRA(a_phi[dit()]),
                       CHF_BOX(m_grids.get(dit())),
                       CHF_CONST_REAL(dx));
      //timercomment//TF_OPERATORLAP_VelocityPoissonOp.stop();
    }
}


bool
VelocityPoissonOp::isDefined() const
{
  // this has to be different from the PoissonOp version
  // because we don't care whether the BC is defined.
  return (m_isDefined);
}
